# -*- coding: utf-8 -*-

import pytest
from mock import MagicMock


@pytest.fixture
def patched_session(monkeypatch):

    monkeypatch.setattr("ctypes.util.find_library", lambda _: "libssh.so.42")
    monkeypatch.setattr("cffi.FFI.dlopen", lambda *_: MagicMock())
    monkeypatch.setattr("pystassh.api.Api.to_string", lambda chars: chars)
    monkeypatch.setattr("pystassh.api.Api.ssh_get_error", lambda *_: "<error message>")

    from pystassh import Session
    return Session


def test_session_init(patched_session):
    session = patched_session("foo", "bar", "baz", "qux", 17)
    assert session._hostname == b'foo'
    assert session._username == b'bar'
    assert session._password == b'baz'
    assert session._passphrase == b'qux'
    assert session._port == b'17'


def test_session_is_connected(monkeypatch, patched_session):

    session = patched_session()
    assert session.is_connected() is False

    session._session = "<session object>"
    monkeypatch.setattr("pystassh.api.Api.ssh_is_connected", lambda *_: 0)
    assert session.is_connected() is False

    monkeypatch.setattr("pystassh.api.Api.ssh_is_connected", lambda *_: 1)
    assert session.is_connected() is True

    session._session = None
    assert session.is_connected() is False


def test_session_connect(monkeypatch, patched_session):

    import pystassh.api
    session = patched_session()
    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: "<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_options_set", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.api.Api.ssh_connect", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.api.Api.ssh_userauth_password", lambda *_: pystassh.api.SSH_AUTH_SUCCESS)
    monkeypatch.setattr("pystassh.api.Api.ssh_userauth_autopubkey", lambda *_: pystassh.api.SSH_AUTH_SUCCESS)
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))

    session.connect()
    assert session._session == "<session object>"

    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: "<new session object>")
    session.connect()
    assert session._session == "<session object>"
    session._session = None

    session.connect()
    assert session._session == "<new session object>"


def test_session_connect_ssh_new_error(monkeypatch, patched_session):

    import pystassh.exceptions
    session = patched_session()
    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: None)
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))

    with pytest.raises(pystassh.exceptions.ConnectionException):
        session.connect()
    assert session._session is None


def test_session_connect_ssh_options_set_error(monkeypatch, patched_session):

    import pystassh.api
    import pystassh.exceptions
    session = patched_session()
    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: "<session object>")
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))

    monkeypatch.setattr("pystassh.api.Api.ssh_options_set",
                        lambda *_: -1 if _[1] == pystassh.api.SSH_OPTIONS_HOST else pystassh.api.SSH_OK)
    with pytest.raises(pystassh.exceptions.ConnectionException):
        session.connect()
    assert session._session is None

    monkeypatch.setattr("pystassh.api.Api.ssh_options_set",
                        lambda *_: -1 if _[1] == pystassh.api.SSH_OPTIONS_PORT_STR else pystassh.api.SSH_OK)
    with pytest.raises(pystassh.exceptions.ConnectionException):
        session.connect()
    assert session._session is None

    monkeypatch.setattr("pystassh.api.Api.ssh_options_set",
                        lambda *_: -1 if _[1] == pystassh.api.SSH_OPTIONS_USER else pystassh.api.SSH_OK)
    with pytest.raises(pystassh.exceptions.ConnectionException):
        session.connect()
    assert session._session is None


def test_session_connect_ssh_connect_error(monkeypatch, patched_session):

    import pystassh.api
    import pystassh.exceptions
    session = patched_session()
    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: "<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_options_set", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))

    monkeypatch.setattr("pystassh.api.Api.ssh_connect", lambda *_: -1)
    with pytest.raises(pystassh.exceptions.ConnectionException):
        session.connect()
    assert session._session is None


def test_session_connect_ssh_userauth_password_error(monkeypatch, patched_session):

    import pystassh.api
    import pystassh.exceptions
    session = patched_session(username="foo", password="bar")
    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: "<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_options_set", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.api.Api.ssh_connect", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))

    monkeypatch.setattr("pystassh.api.Api.ssh_userauth_password", lambda *_: -1)
    with pytest.raises(pystassh.exceptions.AuthenticationException):
        session.connect()
    assert session._session is None


def test_session_connect_ssh_userauth_autopubkey_error(monkeypatch, patched_session):

    import pystassh.api
    import pystassh.exceptions
    session = patched_session()
    monkeypatch.setattr("pystassh.api.Api.ssh_new", lambda *_: "<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_options_set", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.api.Api.ssh_connect", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))

    monkeypatch.setattr("pystassh.api.Api.ssh_userauth_autopubkey", lambda *_: -1)
    with pytest.raises(pystassh.exceptions.AuthenticationException):
        session.connect()
    assert session._session is None


def test_session_disconnect(monkeypatch, patched_session):

    import pystassh.api
    session = patched_session()
    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda *_: False)

    session._session = "<session object>"
    session.disconnect()
    assert session._session is None

    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda *_: True)
    monkeypatch.setattr("pystassh.api.Api.ssh_disconnect", lambda *_: pystassh.api.SSH_OK)
    session._session = "<session object>"
    session.disconnect()
    assert session._session is None


def test_session_with_block(monkeypatch, patched_session):

    def fake_connect(self):
        self._session = "<session object>"

    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))
    monkeypatch.setattr("pystassh.session.Session.connect", fake_connect)
    monkeypatch.setattr("pystassh.session.Session.disconnect", lambda _: None)

    with patched_session() as session:
        assert session.is_connected()


def test_session_execute(monkeypatch, patched_session):

    import pystassh.channel
    import pystassh.exceptions
    session = patched_session()
    with pytest.raises(pystassh.exceptions.PystasshException):
        session.execute("ls")

    monkeypatch.setattr("pystassh.session.Session.is_connected", lambda self: bool(self._session))
    monkeypatch.setattr("pystassh.channel.Channel.execute", lambda _, command: "<result of {}>".format(command))

    session._session = "<session object>"
    session._channel = pystassh.channel.Channel(session._session)
    assert session.execute("ls") == "<result of ls>"


def test_session_get_error_message_error(monkeypatch, patched_session):

    import pystassh.exceptions

    def fake_get_error_message(_):
        raise pystassh.exceptions.UnknownException

    monkeypatch.setattr("pystassh.api.Api.get_error_message", fake_get_error_message)
    session = patched_session()
    assert session.get_error_message() == "<error message irrecoverable>"



# -*- coding: utf-8 -*-

import pytest
from mock import MagicMock


@pytest.fixture
def patched_channel(monkeypatch):

    monkeypatch.setattr("ctypes.util.find_library", lambda _: "libssh.so.42")
    monkeypatch.setattr("cffi.FFI.dlopen", lambda *_: MagicMock())
    monkeypatch.setattr("pystassh.api.Api.to_string", lambda chars: chars)
    monkeypatch.setattr("pystassh.api.Api.ssh_get_error", lambda *_: "<error message>")

    from pystassh.channel import Channel
    return Channel


def test_channel_init(patched_channel):
    channel = patched_channel("<session object>")
    assert channel._session == "<session object>"


def test_channel_is_open(monkeypatch, patched_channel):

    channel = patched_channel("<session object>")
    assert channel._is_open() is False

    channel._channel = "<channel object>"
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_is_open", lambda *_: 0)
    assert channel._is_open() is False

    monkeypatch.setattr("pystassh.api.Api.ssh_channel_is_open", lambda *_: 1)
    assert channel._is_open() is True

    channel._channel = None
    assert channel._is_open() is False


def test_channel_open(monkeypatch, patched_channel):

    import pystassh.api
    channel = patched_channel("<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_new", lambda *_: "<channel object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_open_session", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.channel.Channel._is_open", lambda self: bool(self._channel))

    channel.open()
    assert channel._channel == "<channel object>"

    monkeypatch.setattr("pystassh.api.Api.ssh_channel_new", lambda *_: "<new channel object>")
    channel.open()
    assert channel._channel == "<channel object>"
    channel._channel = None

    channel.open()
    assert channel._channel == "<new channel object>"


def test_channel_open_ssh_channel_new_error(monkeypatch, patched_channel):

    import pystassh.exceptions
    channel = patched_channel("<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_new", lambda *_: None)
    monkeypatch.setattr("pystassh.channel.Channel._is_open", lambda self: bool(self._channel))

    with pytest.raises(pystassh.exceptions.ChannelException):
        channel.open()
    assert channel._channel is None


def test_channel_open_ssh_channel_open_session_error(monkeypatch, patched_channel):

    import pystassh.exceptions
    channel = patched_channel("<session object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_new", lambda *_: "<channel object>")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_open_session", lambda *_: -1)
    monkeypatch.setattr("pystassh.channel.Channel._is_open", lambda self: bool(self._channel))

    with pytest.raises(pystassh.exceptions.ChannelException):
        channel.open()
    assert channel._channel is None


def test_channel_close(monkeypatch, patched_channel):

    import pystassh.api
    channel = patched_channel("<session object>")
    monkeypatch.setattr("pystassh.channel.Channel._is_open", lambda *_: False)

    channel._channel = "<channel object>"
    channel.close()
    assert channel._channel is None

    monkeypatch.setattr("pystassh.channel.Channel._is_open", lambda *_: True)
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_send_eof", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_free", lambda *_: pystassh.api.SSH_OK)
    channel._channel = "<channel object>"
    channel.close()
    assert channel._channel is None


def test_channel_with_block(monkeypatch, patched_channel):

    def fake_open(self):
        self._channel = "<channel object>"

    monkeypatch.setattr("pystassh.channel.Channel._is_open", lambda self: bool(self._channel))
    monkeypatch.setattr("pystassh.channel.Channel.open", fake_open)
    monkeypatch.setattr("pystassh.channel.Channel.close", lambda _: None)

    with patched_channel("<session object>") as channel:
        assert channel._is_open()


def test_channel_execute(monkeypatch, patched_channel):

    import pystassh.api
    import pystassh.exceptions
    import pystassh.result

    def fake_result_init(self, channel, command):
        self._channel = channel
        self._command = command

    monkeypatch.setattr("pystassh.channel.Channel.open", lambda _: None)
    monkeypatch.setattr("pystassh.channel.Channel.close", lambda _: None)
    channel = patched_channel("<sesion object>")

    monkeypatch.setattr("pystassh.api.Api.ssh_channel_request_exec", lambda *_: -1)
    with pytest.raises(pystassh.exceptions.ChannelException):
        channel.execute("ls")

    monkeypatch.setattr("pystassh.api.Api.ssh_channel_request_exec", lambda *_: pystassh.api.SSH_OK)
    monkeypatch.setattr("pystassh.result.Result.__init__", fake_result_init)

    res = channel.execute("ls")
    assert isinstance(res, pystassh.result.Result)
    assert res._command == "ls"
    assert res._channel == channel._channel


def test_channel_get_error_message_error(monkeypatch, patched_channel):

    import pystassh.exceptions

    def fake_get_error_message(_):
        raise pystassh.exceptions.UnknownException

    monkeypatch.setattr("pystassh.api.Api.get_error_message", fake_get_error_message)
    channel = patched_channel("<session object>")
    assert channel.get_error_message() == "<error message irrecoverable>"



# -*- coding: utf-8 -*-

import pytest
from mock import MagicMock


@pytest.fixture
def patched_api(monkeypatch):

    mock = MagicMock()
    del mock.i_dont_exist
    monkeypatch.setattr("ctypes.util.find_library", lambda _: "libssh.so.42")
    monkeypatch.setattr("cffi.FFI.dlopen", lambda *_: mock)

    from pystassh import api
    return api


@pytest.fixture
def patched_init_api(monkeypatch):

    monkeypatch.setattr("ctypes.util.find_library", lambda _: None)
    from pystassh.api import _init_api
    return _init_api


def test_api_getattr(patched_api):

    with pytest.raises(AttributeError):
        patched_api.Api.i_dont_exist()

    assert patched_api.Api.ssh_channel_open_session is patched_api.Api.lib.ssh_channel_open_session


def test_init_api(monkeypatch, patched_api):

    import pystassh.exceptions

    monkeypatch.setattr("ctypes.util.find_library", lambda _: None)

    with pytest.raises(pystassh.exceptions.PystasshException):
        patched_api._init_api()


def test_api_to_string(monkeypatch, patched_api):
    monkeypatch.setattr("pystassh.api.Api.ffi.string", lambda s: s)
    assert patched_api.Api.to_string(b'foo') == b'foo'


def test_api_new_chars(monkeypatch, patched_api):
    monkeypatch.setattr("pystassh.api.Api.ffi.new", lambda s: b'   ')
    assert patched_api.Api.new_chars(3) == b'   '


def test_api_get_error_message(monkeypatch, patched_api):

    import pystassh.exceptions

    def fake_get_error_message(_):
        raise ValueError()

    monkeypatch.setattr("pystassh.api.Api.ssh_get_error", fake_get_error_message)
    with pytest.raises(pystassh.exceptions.UnknownException):
        patched_api.Api.get_error_message("<session object>")


# -*- coding: utf-8 -*-

from . import api


class Result:

    def __init__(self, channel, command):
        """ A Result object contains the execution details of a command.

        Args:
            channel: the libssh's channel instance the result will be attached to
            command: the last command that was run
        """
        self._channel = channel
        self._command = command
        self._buffer_size = 100
        self._stdout = self._read_stdout_or_stderr(False)
        self._stderr = self._read_stdout_or_stderr(True)
        self._return_code = self._read_return_code()

    def __read(self, is_stderr):
        buffer = api.Api.new_chars(self._buffer_size)
        return (api.Api.ssh_channel_read(self._channel, buffer, self._buffer_size, int(is_stderr)),
                api.Api.to_string(buffer))

    def _read_stdout_or_stderr(self, is_stderr):

        content = b""
        count, buffer = self.__read(is_stderr)
        while count:
            content += buffer
            count, buffer = self.__read(is_stderr)
        return content

    def _read_return_code(self):
        return api.Api.ssh_channel_get_exit_status(self._channel)

    @property
    def command(self):
        """ The command from wich the current results came from.
        """
        return self._command

    @property
    def raw_stdout(self):
        """ The raw content of the standard output, as a list of bytes.
        """
        return self._stdout

    @property
    def stdout(self):
        """ The content of the standard output, as a string. Decoding errors are not caught at this level.
        """
        return self.raw_stdout.decode("utf8").rstrip("\r\n")

    @property
    def raw_stderr(self):
        """ The raw content of the standard error output, as a list of bytes.
        """
        return self._stderr

    @property
    def stderr(self):
        """ The content of the standard error output, as a string. Decoding errors are not caught at this level.
        """
        return self.raw_stderr.decode("utf8").rstrip("\r\n")

    @property
    def return_code(self):
        """ The return code of the last command as an int.
        """
        return self._return_code



# -*- coding: utf-8 -*-

""" The Session object is the entry point of any command execution via SSH.

Examples:

    Open a SSH connection, run the "ls" command and print the output.

    >>> with Session('localhost', 'foo', 'bar') as ssh_session:
    ...     result = ssh_session.execute('ls')
    ...     print(result.stdout)

"""

from . import api, exceptions
from . channel import Channel


class Session:

    def __init__(self, hostname="localhost", username="", password="", passphrase="", port=22):

        """ A session object correspond to a unique SSH connexion from which commands can be run.

        Args:
            hostname (str): hostname
            username (str): user name
            password (str): user password
            passphrase (str): optionnal passphrase to be used with a public key authentication
            port (int): SSH remote port
        """
        self._hostname = str.encode(hostname)
        self._username = str.encode(username)
        self._password = str.encode(password)
        self._passphrase = str.encode(passphrase)
        self._port = str.encode(str(port))

        self._session = None
        self._channel = None

    def is_connected(self):
        """ Check if the connexion is currently active.

        Returns:
            bool: A boolean indicating whether or not the connexion is currently active.
        """
        return bool(self._session and api.Api.ssh_is_connected(self._session))

    def connect(self):

        """ Initiate the connection and authentication process on the remote server.

        Raises:
            ConnectionException: if an error occurred during the connection process
            AuthenticationException: if an error occurred during the authentication process
        """
        if self.is_connected():
            return

        session = api.Api.ssh_new()
        if session is None:
            raise exceptions.ConnectionException("Session cannot be created: {}".format(self.get_error_message()))

        ret = api.Api.ssh_options_set(session, api.SSH_OPTIONS_HOST, self._hostname)
        if ret != api.SSH_OK:
            raise exceptions.ConnectionException("Hostname '{}' cannot be set (return code: {}): {}".format(
                self._hostname, ret, self.get_error_message(session)))

        ret = api.Api.ssh_options_set(session, api.SSH_OPTIONS_PORT_STR, self._port)
        if ret != api.SSH_OK:
            raise exceptions.ConnectionException("Port '{}' cannot be set (return code: {}): {}".format(
                self._port, ret, self.get_error_message(session)))

        ret = api.Api.ssh_options_set(session, api.SSH_OPTIONS_USER, self._username)
        if ret != api.SSH_OK:
            raise exceptions.ConnectionException("Username '{}' cannot be set (return code: {}): {}".format(
                self._username, ret, self.get_error_message(session)))

        ret = api.Api.ssh_connect(session)
        if ret != api.SSH_OK:
            raise exceptions.ConnectionException("Connection cannot be made (return code: {}): {}".format(
                ret, self.get_error_message(session)))

        if self._password:
            ret = api.Api.ssh_userauth_password(session, self._username, self._password)
            if ret != api.SSH_AUTH_SUCCESS:
                raise exceptions.AuthenticationException(
                    "Authentication cannot be made with username and password (return code: {}): {}".format(
                        ret, self.get_error_message(session)))
        else:
            ret = api.Api.ssh_userauth_autopubkey(session, self._passphrase)
            if ret != api.SSH_AUTH_SUCCESS:
                raise exceptions.AuthenticationException(
                    "Authentication cannot be made with public key (return code: {}): {}".format(
                        ret, self.get_error_message(session)))

        self._session = session
        self._channel = Channel(self._session)

    def disconnect(self):
        """ Close the current connection.
        """
        if self.is_connected():
            api.Api.ssh_disconnect(self._session)
        self._session = None

    def execute(self, command):
        """ Execute a command on the remote server.

        Args:
            command (str): the command to run

        Returns:
            Result: the Result object for this command
        """
        if not self.is_connected():
            raise exceptions.PystasshException("The session is not ready, call the connect() method first")
        return self._channel.execute(command)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *_):
        self.disconnect()

    def get_error_message(self, session=None):
        """ Tries to retrieve an error message in case of error.

        Args:
            session: a session object that will be used instead of self._session
        Returns:
            str: An error message
        """
        session = session or self._session
        try:
            return api.Api.get_error_message(session)
        except exceptions.UnknownException:
            return "<error message irrecoverable>"



# -*- coding: utf-8 -*-


class PystasshException(Exception):
    """ A base exception for all other Pystassh exceptions.
    """
    pass


class ConnectionException(PystasshException):
    """ Raised when something went wrong during the connection process.
    """
    pass


class AuthenticationException(PystasshException):
    """ Raised when something went wrong during the authentication process.
    """
    pass


class ChannelException(PystasshException):
    """ Raised when an error occurred in the context of a channel.
    """
    pass


class UnknownException(PystasshException):
    """ Sometimes we just don't know what happened...
    """
    pass



# -*- coding: utf-8 -*-

import pytest
from mock import MagicMock


@pytest.fixture
def patched_result(monkeypatch):
    monkeypatch.setattr("ctypes.util.find_library", lambda _: "libssh.so.42")
    monkeypatch.setattr("cffi.FFI.dlopen", lambda *_: MagicMock())
    monkeypatch.setattr("pystassh.api.Api.to_string", lambda chars: chars)
    monkeypatch.setattr("pystassh.api.Api.ssh_get_error", lambda *_: "<error message>")

    from pystassh.result import Result
    return Result


def test_result_init(monkeypatch, patched_result):

    monkeypatch.setattr("pystassh.result.Result._read_stdout_or_stderr", lambda _, b: b"bar\n" if b else b"foo\n")
    monkeypatch.setattr("pystassh.result.Result._read_return_code", lambda _: 0)

    result = patched_result("<channel object>", "ls")
    assert result._channel == "<channel object>"
    assert result._command == "ls"
    assert result._stdout == b"foo\n"
    assert result._stderr == b"bar\n"


def test_result_properties(monkeypatch, patched_result):

    monkeypatch.setattr("pystassh.result.Result._read_stdout_or_stderr", lambda _, b: b"bar\n" if b else b"foo\n")
    monkeypatch.setattr("pystassh.result.Result._read_return_code", lambda _: 17)

    result = patched_result("<channel object>", "ls")
    assert result.stdout == "foo"
    assert result.raw_stdout == b"foo\n"
    assert result.stderr == "bar"
    assert result.raw_stderr == b"bar\n"
    assert result.return_code == 17
    assert result.command == "ls"


def test_result_read_return_code(monkeypatch, patched_result):

    monkeypatch.setattr("pystassh.result.Result._read_stdout_or_stderr", lambda *_: b"")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_get_exit_status", lambda _: 17)

    result = patched_result("<channel object>", "ls")
    assert result._return_code == 17


def test_result_read_stdout_or_stderr(monkeypatch, patched_result):

    def fake_ssh_channel_read(*_):
        if fake_ssh_channel_read.nb_loops == 3:
            return 0
        fake_ssh_channel_read.nb_loops += 1
        return 3
    fake_ssh_channel_read.nb_loops = 0

    monkeypatch.setattr("pystassh.result.Result._read_return_code", lambda _: 17)
    monkeypatch.setattr("pystassh.api.Api.new_chars", lambda _: b"foo")
    monkeypatch.setattr("pystassh.api.Api.ssh_channel_read", fake_ssh_channel_read)

    result = patched_result("<channel object>", "ls")
    assert result._stdout == b"foofoofoo"



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name="pystassh",
    version="1.0",
    author="Julien Chaumont",
    author_email="pystassh@julienc.io",
    description="An easy to use libssh wrapper to execute commands on a remote server via SSH with Python",
    license="MIT",
    url="http://github.com/julienc91/pystassh/",
    packages=find_packages(),
    install_requires=["cffi"],
    setup_requires=['pytest-runner'],
    tests_require=["pytest", "pytest-cov", "mock"]
)



# -*- coding: utf-8 -*-

import ctypes.util
from cffi import FFI
from . import exceptions


SSH_OK = 0
SSH_AUTH_SUCCESS = 0

SSH_OPTIONS_HOST = 0
SSH_OPTIONS_PORT_STR = 2
SSH_OPTIONS_USER = 4


def _init_api():

    lib_name = ctypes.util.find_library("ssh")
    if not lib_name:
        raise exceptions.PystasshException("libssh not found, please visit https://www.libssh.org/get-it/")

    ffi = FFI()
    lib = ffi.dlopen(lib_name)
    ffi.cdef("""
        void* ssh_new();
        int ssh_options_set(void*, int, char*);
        int ssh_connect(void*);
        int ssh_disconnect(void*);
        int ssh_is_connected(void*);
        char* ssh_get_error(void*);

        int ssh_userauth_password(void*, char*, char*);
        int ssh_userauth_autopubkey(void*, char*);

        void* ssh_channel_new(void*);
        int ssh_channel_open_session(void*);
        int ssh_channel_is_open(void*);
        void ssh_channel_free(void*);
        int ssh_channel_request_exec(void*, char*);
        int ssh_channel_request_pty(void*);
        int ssh_channel_request_shell(void*);

        int ssh_channel_get_exit_status(void*);
        int ssh_channel_read(void*, char*, int, int);
        int ssh_channel_send_eof(void*);
    """)
    return ffi, lib


class ApiMetaclass(type):

    def __getattr__(cls, name):
        if hasattr(cls.lib, name):
            return getattr(cls.lib, name)
        raise AttributeError("Api has no attribute {}".format(name))


class Api(metaclass=ApiMetaclass):

    ffi, lib = _init_api()

    @classmethod
    def to_string(cls, chars):
        return cls.ffi.string(chars)

    @classmethod
    def new_chars(cls, size):
        return cls.ffi.new("char[{}]".format(size))

    @classmethod
    def get_error_message(cls, session):
        try:
            raw_error = cls.ssh_get_error(session)
        except (TypeError, ValueError, AttributeError):
            raise exceptions.UnknownException("Error message irrecoverable")
        return cls.to_string(raw_error)



# -*- coding: utf-8 -*-

from . import api, exceptions
from . result import Result


class Channel:

    def __init__(self, session):
        """ A channel is an environment bound to a session in which commands can be run.

        Args:
            session: the libssh's session instance the channel will be bound to
        """
        self._session = session
        self._channel = None
        self._stdout = None
        self._stderr = None

    def _is_open(self):
        return bool(self._channel and api.Api.ssh_channel_is_open(self._channel))

    def open(self):
        """ Open a new channel.

        Raises:
            ChannelException: if the channel could not be correctly initialized
        """
        if self._is_open():
            return

        channel = api.Api.ssh_channel_new(self._session)
        if channel is None:
            raise exceptions.ChannelException("Channel cannot be created: {}".format(self.get_error_message()))

        ret = api.Api.ssh_channel_open_session(channel)
        if ret != api.SSH_OK:
            raise exceptions.ChannelException("Channel cannot be opened: {}".format(self.get_error_message()))

        self._channel = channel

    def close(self):
        """ Close the current channel.
        """
        if self._is_open():
            api.Api.ssh_channel_send_eof(self._channel)
            api.Api.ssh_channel_free(self._channel)
        self._channel = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *_):
        self.close()

    def execute(self, command):
        """ Execute a command.

        Args:
            command (str): the command to run

        Returns:
            Result: the Result object for this command
        """
        with self:
            ret = api.Api.ssh_channel_request_exec(self._channel, str.encode(command))
            if ret != api.SSH_OK:
                raise exceptions.ChannelException("Command cannot be executed (return code: {}): {}".format(
                    command, self.get_error_message()))
            return Result(self._channel, command)

    def get_error_message(self):
        """ Tries to retrieve an error message in case of error.

        Returns:
            str: An error message
        """
        try:
            return api.Api.get_error_message(self._session)
        except exceptions.UnknownException:
            return "<error message irrecoverable>"
